/* -*- Mode: C; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 8 -*-
 *
 * msd-ldsm-trash-empty.h
 * Copyright (C) Chris Coulson 2009 <chrisccoulson@googlemail.com>
 * 
 * msd-ldsm-trash-empty.h is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * msd-ldsm-trash-empty.h is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef _msd_ldsm_trash_empty_h_
#define _msd_ldsm_trash_empty_h_

#include <gtk/gtk.h>

void msd_ldsm_trash_empty (void);

#endif /* _msd_ldsm_trash_empty_h_ */ 
