/*
 * Copyright © 2001 Jonathan Blandford <jrb@gnome.org>
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Red Hat not be used in advertising or
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.  Red Hat makes no representations about the
 * suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * Authors:  Jonathan Blandford
 */

#ifndef LOCATE_POINTER_H
#define LOCATE_POINTER_H

#include <gdk/gdk.h>

void msd_locate_pointer (GdkDisplay *display);

#endif
